#include <Sockets/Exceptions.hpp>
#include <Sockets/TcpSocket_Old.hpp>
#include <Sockets/Sockets.hpp>
#include <sys/types.h>

#include <cerrno>

#include <cstring>
#include <memory>
#include <string>
#include <iostream>

#include <jlog/Logger.hpp>

#if defined(__linux__) || defined(__APPLE__)
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#elif defined(WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

namespace Socket {

	TcpSocket_Old::TcpSocket_Old()
	{
		mInfo = new addrinfo;
		memset(mInfo, 0, sizeof *mInfo);
		mInfo->ai_family = AF_UNSPEC;
		mInfo->ai_socktype = SOCK_STREAM;
	}

	TcpSocket_Old::TcpSocket_Old(int family, int flags)
	{
		mInfo = new addrinfo;
		memset(mInfo, 0, sizeof *mInfo);
		mInfo->ai_family = family;
		mInfo->ai_socktype = SOCK_STREAM;
		mInfo->ai_flags = flags;

		if (family == AF_UNSPEC)
			return; // Can't create socket now


		mSock = ::socket(mInfo->ai_family, mInfo->ai_socktype, 0);
		if (mSock == -1)
        {
            auto errcode = std::string(strerror(errno));

        }
			throw SocketCreationException(strerror(errno));
		// Socket successfully "opened"
		mSockCreated = true;
	}

	TcpSocket_Old::TcpSocket_Old(int socket, addrinfo info, bool bound, bool connected)
		: mSock(socket), mBound(bound), mConnected(connected)
	{
		mInfo = new addrinfo(info);
	}

	TcpSocket_Old::~TcpSocket_Old()
	{
		if (mSockCreated && (mClosed == false)) // Little more clear to outright say "false"
			Close();
		freeaddrinfo(mInfo);
	}

	void TcpSocket_Old::Bind(uint16_t port)
	{
		if (mBound && mConnected)
			throw SocketBindingException("Already bound!");

		setInfo(port);

		addrinfo * result;

		for (result = mInfo; result != NULL; result = mInfo->ai_next)
		{
			if (!mSockCreated)
			{
				try {
					openSocket(result);
				} catch(SocketCreationException &e) {
					continue;
				}
			}

			// Socket successfully opened from here
            int ret = ::bind(mSock, result->ai_addr, result->ai_addrlen);
			if (ret < 0)
                CheckErrors_bind(errno);
            mBound = true;
            return;
		}
		// Couldn't bind, throw
		throw SocketBindingException("Can't bind to port");
	}

    void TcpSocket_Old::BindAny() {
        Bind(INPORT_ANY);
    }
    void TcpSocket_Old::BindAny(uint16_t& port)
    {
        this->Bind(INPORT_ANY);
        //portno = port;
    }


    void TcpSocket_Old::Connect(const IPAddress &ipaddr) {
        Connect(ipaddr.addr_string(), ipaddr.port);
    }

	void TcpSocket_Old::Connect(std::string address, uint16_t port)
	{
		if (mConnected)
			throw SocketConnectException("Already Connected!");

		setInfo(address, htons(port));

		addrinfo * result;
		for (result = mInfo; result!= NULL; result = mInfo->ai_next)
		{
			if (!mSockCreated) {
				try {
					openSocket(result);
				} catch (SocketCreationException &e) {
					continue;
				}
			}
			// Socket successfully opened from here
            int status = 0;

            status = ::connect(mSock, result->ai_addr, result->ai_addrlen) == 0;
			if (status < 0)
				CheckErrors_connect(errno);
            mConnected = true;
            return;
		}
	}

	void TcpSocket_Old::Listen(int maxQueue)
	{
		if (::listen(mSock, maxQueue) != 0)
            CheckErrors_accept(errno);

		jlog::Debug("Listening...");
	}



	void TcpSocket_Old::Send(const u8 *data, unsigned int length, int flags)
	{
		const u8 * buff = data;
		int status = 0;
		int total_sent = 0;
		unsigned int left_to_send = length;
		while(total_sent < length) {
			status = ::send(mSock, reinterpret_cast<const char*>(buff + total_sent), left_to_send, flags);
			if (status == -1)
                CheckErrors_send(errno);
			else {
				total_sent += status;
				left_to_send -= status;
			}
		}
	}

	bool TcpSocket_Old::Receive(u8* msg, int len, int flags)
	{
		int status = 0;
		if ((status = ::recv(mSock, reinterpret_cast<char*>(msg), len, flags)) == -1)
            CheckErrors_recv(errno);
        else if (status == 0)
			return false;

		return true;
	}

	void TcpSocket_Old::Close()
	{
#if defined(__linux__) || defined(__APPLE__)
        //::shutdown(mSock, 2);

		if (::close(mSock) == -1)
            CheckErrors_close(errno);
		else
			mClosed = true;
#elif defined(_WIN32)
	if (::closesocket(mSock == -1))
		CheckErrors_close(errno);
	else
		mClosed = true;
#endif
	}
	// Private:
	void TcpSocket_Old::setInfo(int port)
	{
		setInfo("null", port);
	}
	void TcpSocket_Old::setInfo(std::string address, int port)
	{
		const char *charAddress;
		if (address == "null")
			charAddress = NULL;
		else
			charAddress = address.c_str();

		addrinfo hints = *mInfo;
		int status = getaddrinfo(charAddress, std::to_string(port).c_str(), &hints, &mInfo);
		if ((status) != 0)
		{
			delete charAddress;

            CheckErrors_getaddrinfo(errno);

			throw SocketException(
			("getaddrinfo returned non-zero: " + std::string(gai_strerror(status))).c_str()
			);			
		}
		delete charAddress;
	}

	void TcpSocket_Old::openSocket(addrinfo * info)
	{
		mSock = ::socket(info->ai_family, info->ai_socktype, info->ai_protocol);
		if (mSock == -1)
		{
            CheckErrors_socket(errno);
		}
	}

    std::shared_ptr<TcpSocket_Old> TcpSocket_Old::Accept() {
        jlog::Debug("Starting to accept");
        union {
            sockaddr addr;
            sockaddr_in in;
            sockaddr_in6 in6;
            sockaddr_storage s;
        } address;
        socklen_t addressSize = sizeof(sockaddr_storage);
        int newSock;
#if defined(__linux__) || defined(__APPLE__)
        if ((newSock = ::accept(mSock, (struct sockaddr *) 0, (unsigned int *) 0)) == -1) {
            //if ((newSock = ::accept(mSock, (sockaddr*)&address.s, &addressSize)) == -1) {
            jlog::Debug(strerror(errno));
            CheckErrors_accept(errno);
        }
#elif defined(_WIN32)
		if ((newSock = ::accept(mSock, (struct sockaddr *) 0, (int *) 0)) == -1) {
			//if ((newSock = ::accept(mSock, (sockaddr*)&address.s, &addressSize)) == -1) {
			jlog::Debug(strerror(errno));
			CheckErrors_accept(errno);
		}
#endif
        jlog::Debug("1 client accepted");

        addrinfo info;
        memset(&info, 0, sizeof info);
        if (address.s.ss_family == AF_INET) {
            info.ai_family = AF_INET;
            info.ai_addr = new sockaddr(address.addr);
        } else {
            info.ai_family = AF_INET6;
            info.ai_addr = new sockaddr(address.addr);
        }
        return std::make_shared<TcpSocket_Old>(newSock, info, true, false);

    }
}
